import numpy as np
import traceback
from importlib.util import find_spec
import itertools
from scipy.optimize import curve_fit
from scipy.special import erf

try:
    import matplotlib.pyplot as plt
    import seaborn as sns
except ModuleNotFoundError:
    raise RuntimeError(
        "ROC estimation can only be done when matplotlib and seaborn are installed."
    )
sns.set(context="talk", style="ticks")


class ScoreHistogramFdr:
    def __init__(self):
        self.fig = plt.figure(figsize=(5 * 2, 5))
        self.hist_ax = self.fig.add_subplot(1, 2, 1)
        self.fdr_ax = self.fig.add_subplot(1, 2, 2)

    def draw_histogram(self, scores, nbins=30, return_bins=False):
        y, x_hist, _ = self.hist_ax.hist(
            scores, bins=nbins, histtype="step", color="grey"
        )
        self.hist_ax.set_xlabel(r"${LCC}_{max}$")
        self.hist_ax.set_xlim(x_hist[0], x_hist[-1])
        self.hist_ax.set_ylabel("Frequency")
        if return_bins:
            return y, x_hist

    def draw_bimodal(self, x, y1, y2, ymax=None):
        # plot bimodal model and the gaussian particle population
        self.hist_ax.plot(
            x, y1, lw=3.5, alpha=0.9, color="tab:blue"
        )  # , label='Bimodal model')
        self.hist_ax.plot(
            x, y2, lw=4, alpha=0.9, color="tab:orange"
        )  # , label='True positives')
        # population = params[2:5]
        if ymax is not None:
            self.hist_ax.set_ylim(0, ymax)
        # self.hist_ax.legend(loc='upper right')

    def draw_score_threshold(self, x, ymax):
        self.hist_ax.vlines(
            x, 0, ymax, linestyle="dashed", label=f"Cutoff: {x:.2f}", color="black"
        )
        self.hist_ax.legend(loc="upper right")

    def draw_fdr_recall(self, fdr, recall, optimal_id, ruc):
        self.fdr_ax.scatter(fdr, recall, facecolors="none", edgecolors="gray", s=25)
        # add optimal threshold in green
        self.fdr_ax.scatter(
            fdr[optimal_id],
            recall[optimal_id],
            s=25,
            color="black",
            label=f"RUC: {ruc:.2f}",
        )
        self.fdr_ax.plot([0, 1], [0, 1], ls="--", c=".3", lw=1)
        self.fdr_ax.set_xlabel("FDR")
        self.fdr_ax.set_ylabel("Recall")
        self.fdr_ax.set_xlim(0, 1)
        self.fdr_ax.set_xticks([0, 0.2, 0.4, 0.6, 0.8, 1])
        self.fdr_ax.set_ylim(0, 1)
        self.fdr_ax.set_yticks([0, 0.2, 0.4, 0.6, 0.8, 1])
        self.fdr_ax.legend(loc="lower right")

    def write(self, filename, quality=300, transparency=False, bbox="tight"):
        plt.tight_layout()
        plt.savefig(filename, dpi=quality, transparent=transparency, bbox_inches=bbox)

    def display(self):
        plt.tight_layout()
        plt.show()


def check_square_fdr(fdr, recall, epsilon=1e-3):
    """
    @param fdr: list of fdr values
    @type  fdr: L{list}
    @param recall: list of recall values
    @type  recall: L{list}
    @param epsilon: tolerance for closeness to 0 and 1
    @type  epsilon: L{float}

    @return: boolean of whether the FDR is almost square within tolerance
    @rtype:  L{bool}

    @author: Marten Chaillet
    """

    # fdr and recall should contain values very close to 0 and 1, respectively
    # if function is square
    union = [
        (f, r)
        for f, r in zip(fdr, recall)
        if ((np.abs(0.0 - f) < epsilon) and (np.abs(1.0 - r) < epsilon))
    ]
    return bool(union)


def distance_to_diag(fdr, recall):
    """
    @param fdr: list of fdr values
    @type  fdr: L{list}
    @param recall: list of recall values
    @type  recall: L{list}

    @return: list of distance of each fdr, recall combination to diagonal line
    @rtype:  L{list}

    @author: Marten Chaillet
    """
    # two point on the diagonal to find the distance to
    lp1, lp2 = (0, 0), (1, 1)
    # list to hold distances
    distance = []
    for f, r in zip(fdr, recall):
        d = np.abs(
            (lp2[0] - lp1[0]) * (lp1[1] - r) - (lp1[0] - f) * (lp2[1] - lp1[1])
        ) / np.sqrt((lp2[0] - lp1[0]) ** 2 + (lp2[1] - lp1[1]) ** 2)
        distance.append(d)
    return distance


def calculate_histogram(scores, num_steps):
    # construct x and y array according to the given peak index
    # preferably input is already sorted
    scores.sort()  # this is sorted from lowest to highest
    min = scores[0]
    max = scores[-1]

    step = (max - min) / num_steps
    x = []
    for i in range(num_steps):
        x.append(min + i * step)
    x.append(max)

    y = []
    for i in range(num_steps):
        lower = x[i]
        upper = x[i + 1]
        n = len([v for v in scores if lower <= v <= upper])
        y.append(n)

    return x, y


def evaluate_estimates(estimated_positions, ground_truth_positions, tolerance):
    """
    Estimated_positions numpy array, ground truth positions numpy array
    :param estimated_positions:
    :type estimated_positions:
    :param ground_truth_positions:
    :type ground_truth_positions:
    :param tolerance:
    :type tolerance:
    :return:
    :rtype:
    """
    from scipy.spatial.distance import cdist

    n_estimates = estimated_positions.shape[0]
    matrix = cdist(estimated_positions, ground_truth_positions, metric="euclidean")
    correct = [0] * n_estimates
    for i in range(n_estimates):
        if matrix[i].min() < tolerance:
            correct[i] = 1
    return correct


def fdr_recall(correct_particles, scores):
    assert all(i > j for i, j in itertools.pairwise(scores)), print(
        "Scores list should be decreasing."
    )

    n_true_positives = sum(correct_particles)
    true_positives, false_positives = 0, 0
    fdr, recall = [], []
    for correct, score in zip(correct_particles, scores):
        if correct:
            true_positives += 1
        else:
            false_positives += 1
        if n_true_positives == 0:
            recall.append(0)
        else:
            recall.append(true_positives / n_true_positives)
        fdr.append(false_positives / (true_positives + false_positives))
    return fdr, recall


def get_distance(line, point):
    a1, b1 = line
    x, y = point
    a2 = -(1 / a1)
    b2 = y - a2 * x

    x_int = (b2 - b1) / (a1 - a2)
    y_int = a2 * x_int + b2

    return np.sqrt((x_int - x) ** 2 + (y_int - y) ** 2)


def distance_to_random(fdr, recall):
    auc = [0] * len(fdr)
    for i in range(len(fdr)):
        d = get_distance((1, 0), (fdr[i], recall[i]))  # AUC should be 1 at most not 1/2
        if recall[i] > fdr[i]:
            auc[i] = d
        else:
            auc[i] = -d
    return max(auc), np.argmax(auc)


# ========== functions for fitting ==========
# define gaussian function with parameters to fit
def gauss(x, mu, sigma, amp):
    return amp * np.exp(-((x - mu) ** 2) / (2 * sigma**2))


# integral of gaussian with certain sigma and A
def gauss_integral(sigma, amp):
    # mu does not influence the integral
    return amp * np.abs(sigma) * np.sqrt(2 * np.pi)


# define bimodal function of two gaussians to fit both populations
def bimodal(x, mu1, sigma1, amp1, mu2, sigma2, amp2):
    return gauss(x, mu1, sigma1, amp1) + gauss(x, mu2, sigma2, amp2)


def plist_quality_gaussian_fit(
    lcc_max_values,
    score_volume,
    particle_peak_index,
    force_peak=False,
    output_figure_name=None,
    crop_hist=False,
    num_bins=30,
    n_tomograms=1,
):
    # read out the scores
    correlation_scores = np.array(sorted(lcc_max_values, reverse=True))

    # draw the histogram
    plot = ScoreHistogramFdr()
    y, x_hist = plot.draw_histogram(
        correlation_scores, nbins=num_bins, return_bins=True
    )

    try:
        # ===== fit bimodal distribution =====
        # adjust x to center of each bin so len(x)==len(y)
        x = (x_hist[1:] + x_hist[:-1]) / 2
        hist_step = x_hist[1] - x_hist[0]

        # noise gaussian std
        # noise_sigma = np.sqrt((score_volume.std() ** 2) * n_tomograms)
        # if n_tomograms > 1 else score_volume.std()
        noise_sigma = score_volume.std()
        noise_mean = score_volume.mean()
        noise_size = score_volume.size * n_tomograms

        # noise gaussian A value
        noise_a = ((noise_size) / (noise_sigma * np.sqrt(2 * np.pi))) * hist_step

        # expected values
        # left gaussian expectation:
        #     score volume is skewed gaussian, fit only sigma with upper limit
        #     (skewed because it only contains highest score at each position)
        # right gaussian expectation:
        #     mu ~ x[half] and A ~ y[half]
        expected = (noise_sigma, x[particle_peak_index], 0.1, y[particle_peak_index])
        # force peak of particle population to be at peak index
        if force_peak:
            bounds = (
                [noise_sigma, x[particle_peak_index] - 0.01, 0, 0],
                [noise_sigma * 1.5, x[particle_peak_index] + 0.01, 0.1, y[1]],
            )
        else:
            bounds = (
                [noise_sigma, x[int(len(x) * 0.25)], 0, 0],
                [noise_sigma * 1.5, x[-1], 0.1, y[1]],
            )

        # TODO use lambda expression to fix mu_1 and sigma_1
        # params_names = ['mu_1', 'sigma_1', 'A_1', 'mu_2', 'sigma_2', 'A_2']
        params_names = ["sigma_1", "mu_2", "sigma_2", "A_2"]
        # skip first position as the noise peak there is likely incorrect
        params, cov = curve_fit(
            lambda x, p1, p2, p3, p4: bimodal(x, noise_mean, p1, noise_a, p2, p3, p4),
            x[1:],
            y[1:],
            p0=expected,
            bounds=bounds,
            maxfev=2000,
        )  # max iterations argument: maxfev=2000)
        # give sigma of fit for each parameter
        sigma = np.sqrt(np.diag(cov))

        # print information about fit of the model
        print("\nfit of the bimodal model:")
        print("\testimated\t\tsigma")
        for n, p, s in zip(params_names, params, sigma):
            print(f"{n}\t{p:.3f}\t\t{s:.3f}")
        print("\n")

        noise, population = ((noise_mean, params[0], noise_a), tuple(params[1:4]))
        y_bimodal = bimodal(x, *noise, *population)
        y_gauss = gauss(x, *population)

        if crop_hist:
            plot.draw_bimodal(x, y_bimodal, y_gauss, ymax=3 * population[2])
        else:
            plot.draw_bimodal(x, y_bimodal, y_gauss)

        # ===== Generate a ROC curve =====
        roc_steps = 50
        x_roc = np.flip(np.linspace(x[0], x[-1], roc_steps))

        # find ratio of hist step vs roc step
        roc_step = (x[-1] - x[0]) / roc_steps
        delta = (
            hist_step / roc_step
        )  # can be used to divide true/false positives by per roc step
        # variable for total number of tp and fp
        n_false_positives = 0.0
        # list for storing probability of true positives and false positives
        # for each cutoff
        recall = (
            []
        )  # recall = TP / (TP + FN); TP + FN is the full area under the Gaussian curve
        fdr = []  # false discovery rate = FP / (TP + FP); == 1 - TP / (TP + FP)

        # find integral of gaussian particle population;
        # NEED TO DIVIDE BY HISTOGRAM BIN STEP
        population_integral = gauss_integral(population[1], population[2]) / hist_step
        print(
            f" - estimation total number of true positives: {population_integral:.1f}"
        )

        # should use CDF (cumulative distribution function) of Gaussian,
        # gives probability from -infinity to x
        def cdf(x):
            return 0.5 * (1 + erf((x - population[0]) / (np.sqrt(2) * population[1])))

        def gauss_noise(x):
            return gauss(x, *noise)

        for x_i in x_roc:
            # calculate probability of true positives x_i
            # n_true_positives += gauss_pop(x_i) / delta
            n_true_positives = (1 - cdf(x_i)) * population_integral

            # determine false positives up to this point, could also use CDF
            n_false_positives += gauss_noise(x_i) / delta

            # add probability
            recall.append(n_true_positives / population_integral)
            fdr.append(n_false_positives / (n_true_positives + n_false_positives))

        # find best classifier by calculating the rectangle under the curve
        # for each roc point
        recall = np.array(recall)
        fdr = np.array(fdr)
        rectangles = recall * (1 - fdr)
        cutoff, ruc = rectangles.argmax(), rectangles.max()

        # plot the threshold on the distribution plot for visual inspection
        plot.draw_score_threshold(x_roc[cutoff], max(y))
        print(f" - optimal correlation coefficient threshold is {x_roc[cutoff]:.3f}")
        print(
            (
                " - this threshold approximately selects "
                f"{(1 - cdf(x_roc[cutoff])) * population_integral:.1f} particles",
            )
        )

        # plot the fdr curve
        plot.draw_fdr_recall(fdr, recall, cutoff, ruc)
        print("Rectangle Under Curve (RUC): ", ruc)

    except (RuntimeError, ValueError):
        # runtime error is because the model could not be fit,
        # in that case print error and continue with execution
        traceback.print_exc()

    if output_figure_name is None:
        plot.display()
    else:
        if output_figure_name.suffix not in [".svg", ".png"]:
            output_figure_name = output_figure_name + ".png"
        plot.write(output_figure_name)



import numpy.typing as npt
import numpy as np
import voltools as vt
import logging
from scipy.ndimage import center_of_mass, zoom
from scipy.fft import rfftn, irfftn
from typing import Optional
from pytom_tm.weights import create_ctf, create_gaussian_low_pass, radial_average

plotting_available = True
try:
    import matplotlib.pyplot as plt
    import seaborn as sns
    sns.set(context='talk', style='ticks')
except ModuleNotFoundError:
    plotting_available = False


def generate_template_from_map(
        input_map: npt.NDArray[float],
        input_spacing: float,
        output_spacing: float,
        ctf_params: Optional[dict] = None,
        center: bool = False,
        filter_to_resolution: Optional[float] = None,
        output_box_size: Optional[int] = None,
        display_filter: bool = False
) -> npt.NDArray[float]:
    """Generate a template from a density map.

    Parameters
    ----------
    input_map: npt.NDArray[float]
        3D density to use for generating the template, if box is not square it will be padded to square
    input_spacing: float
        voxel size of input map (in A)
    output_spacing: float
        voxel size of output map (in A) the ratio of input to output will be used for downsampling
    ctf_params: Optional[dict], default None
        dictionary with CTF params to apply to the template, see pytom_tm.weights.create_ctf() for details
    center: bool, default False
        set to True to center the template in the box by calculating the center of mass
    filter_to_resolution: Optional[float], default None
        low-pass filter resolution to apply to template, if not provided will be set to 2 * output_spacing
    output_box_size:  Optional[int], default None
        final box size of template
    display_filter: bool, default False
        flag to display a plot of the filter applied to the template

    Returns
    -------
    template: npt.NDArray[float]
        processed template in the specified output box size, box will be square
    """
    # make the map a box with equal dimensions
    if len(set(input_map.shape)) != 1:
        diff = [max(input_map.shape) - s for s in input_map.shape]
        input_map = np.pad(
            input_map,
            tuple([(d // 2, d // 2 + d % 2) for d in diff]),
            mode='constant',
            constant_values=0,
        )

    if filter_to_resolution is None:
        # Set to nyquist resolution
        filter_to_resolution = 2 * output_spacing
    elif filter_to_resolution < (2 * output_spacing):
        warning_text = (f"Filter resolution is too low,"
                        f" setting to {2 * output_spacing}A (2 * output voxel size)")
        logging.warning(warning_text)
        filter_to_resolution = 2 * output_spacing

    # extend volume to the desired output size before applying convolutions!
    if output_box_size is not None:
        logging.debug(f'size check {output_box_size} > {(input_map.shape[0] * input_spacing) // output_spacing}')
        if output_box_size > (input_map.shape[0] * input_spacing) // output_spacing:
            pad = int(output_box_size * (output_spacing / input_spacing)) - input_map.shape[0]
            logging.debug(f'pad with this number of zeros: {pad}')
            input_map = np.pad(
                input_map,
                (pad // 2, pad // 2 + pad % 2),
                mode='constant',
                constant_values=0
            )
        elif output_box_size < (input_map.shape[0] * input_spacing) // output_spacing:
            logging.warning('Could not set specified box size as the map would need to be cut and this might '
                            'result in loss of information of the structure. Please decrease the box size of the map '
                            'by hand (e.g. chimera)')

    if center:
        volume_center = np.divide(np.subtract(input_map.shape, 1), 2, dtype=np.float32)
        input_center_of_mass = center_of_mass(input_map)
        shift = np.subtract(volume_center, input_center_of_mass)
        input_map = vt.transform(input_map, translation=shift, device='cpu')

        logging.debug(f'center of mass, before was '
                      f'{np.round(input_center_of_mass, 2)} '
                      f'and after {np.round(center_of_mass(input_map), 2)}')

    # create ctf function and low pass gaussian if desired
    # for ctf the spacing of pixels/voxels needs to be in meters (not angstrom)
    ctf = 1 if ctf_params is None else create_ctf(input_map.shape, **ctf_params).astype(np.float32)
    lpf = create_gaussian_low_pass(input_map.shape, input_spacing, filter_to_resolution).astype(np.float32)

    if display_filter and plotting_available:
        q, average = radial_average(ctf * lpf)
        fig, ax = plt.subplots()
        ax.plot(q / len(q), average)
        ax.set_xlabel('Fraction of Nyquist')
        ax.set_ylabel('Contrast transfer')
        plt.show()
    elif display_filter and not plotting_available:
        logging.info('Plotting not possible as matplotlib is not installed')

    logging.info('Convoluting volume with filter and then downsampling.')
    return zoom(
        irfftn(rfftn(input_map) * ctf * lpf),
        input_spacing / output_spacing
    )



import logging
from importlib import metadata
__version__ = metadata.version('pytom-match-pick')


try:
    import cupy
except (ModuleNotFoundError, ImportError):
    logging.warning('Error for template matching: cupy installation not found or not functional.')



import unittest
import pathlib
from pytom_tm.angles import load_angle_list, AVAILABLE_ROTATIONAL_SAMPLING


TEST_DATA_DIR = pathlib.Path(__file__).parent.joinpath('test_data')
ERRONEOUS_ANGLE_FILE = TEST_DATA_DIR.joinpath('error_angles.txt')


class TestAngles(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        TEST_DATA_DIR.mkdir()
        # create an erroneous angle file
        with open(ERRONEOUS_ANGLE_FILE, 'w') as fstream:
            fstream.write(' '.join(map(str, [1.] * 4)))
            fstream.write(' '.join(map(str, [1.] * 3)))

    @classmethod
    def tearDownClass(cls) -> None:
        ERRONEOUS_ANGLE_FILE.unlink()
        TEST_DATA_DIR.rmdir()

    def test_load_list(self):
        with self.assertRaises(ValueError, msg='Invalid angle file should raise an error'):
            load_angle_list(ERRONEOUS_ANGLE_FILE)

        for sampling, (angle_file, n_rotations) in AVAILABLE_ROTATIONAL_SAMPLING.items():
            angle_list = load_angle_list(angle_file)
            self.assertEqual(len(angle_list), n_rotations, msg=f'Unexpected number of rotations for {angle_file}')



import unittest
import numpy as np
from pytom_tm.template import generate_template_from_map

class TestLogging(unittest.TestCase):
    def test_lowpass_resolution(self):
        template = np.zeros((13, 13, 13), dtype=np.float32)

        # Test too low filter resolution
        with self.assertLogs(level='WARNING') as cm:
            _ = generate_template_from_map(template, 1., 1., filter_to_resolution=1.5)
        self.assertEqual(len(cm.output), 1)
        self.assertIn('Filter resolution', cm.output[0])
        self.assertIn('too low', cm.output[0])

        # Test working filter resolution
        with self.assertNoLogs(level='WARNING'):
            _ = generate_template_from_map(template, 1., 1., filter_to_resolution=2.5)
 



import unittest
import voltools as vt
import cupy as cp
from importlib_resources import files
from pytom_tm.mask import spherical_mask
from pytom_tm.angles import load_angle_list
from pytom_tm.correlation import normalised_cross_correlation


class TestMask(unittest.TestCase):
    def setUp(self):
        self.angles = load_angle_list(files('pytom_tm.angle_lists').joinpath('angles_50.00_100.txt'))

    def test_rotational_invariance_even(self):
        # TEST EVEN MASK
        nxcc_offcenter, nxcc_centered = [], []

        mask = cp.asarray(spherical_mask(12, 4, 0.5))
        mask_rotated = cp.zeros_like(mask)

        mask_texture = vt.StaticVolume(
            mask,
            interpolation='filt_bspline',
            device='gpu'
        )

        for i in range(len(self.angles)):

            mask_texture.transform(
                rotation=self.angles[i],
                rotation_units='rad',
                rotation_order='rzxz',
                center=tuple([x // 2 for x in mask.shape]),
                output=mask_rotated
            )

            nxcc_offcenter.append(normalised_cross_correlation(mask, mask_rotated).get())

        for i in range(len(self.angles)):

            mask_texture.transform(
                rotation=self.angles[i],
                rotation_units='rad',
                rotation_order='rzxz',
                output=mask_rotated,
                # center=np.divide(np.subtract(mask.shape, 1), 2, dtype=np.float32),
            )

            nxcc_centered.append(normalised_cross_correlation(mask, mask_rotated).get())

        self.assertTrue(sum(nxcc_centered) > sum(nxcc_offcenter),
                        msg='Center of rotation for mask is incorrect.')
        self.assertTrue(sum(nxcc_centered) > 99.27, msg='Precision of mask rotation is too low.')

    def test_rotational_invariance_uneven(self):
        # TEST UNEVEN MASK
        nxcc_offcenter, nxcc_centered = [], []

        mask = cp.asarray(spherical_mask(13, 4, 0.5))
        mask_rotated = cp.zeros_like(mask)

        mask_texture = vt.StaticVolume(
            mask,
            interpolation='filt_bspline',
            device='gpu'
        )

        for i in range(len(self.angles)):
            mask_texture.transform(
                rotation=self.angles[i],
                rotation_units='rad',
                rotation_order='rzxz',
                center=tuple([x // 2 for x in mask.shape]),
                output=mask_rotated
            )

            nxcc_offcenter.append(normalised_cross_correlation(mask, mask_rotated).get())

        for i in range(len(self.angles)):
            mask_texture.transform(
                rotation=self.angles[i],
                rotation_units='rad',
                rotation_order='rzxz',
                output=mask_rotated,
                # center=np.divide(np.subtract(mask.shape, 1), 2, dtype=np.float32),
            )

            nxcc_centered.append(normalised_cross_correlation(mask, mask_rotated).get())

        self.assertAlmostEqual(sum(nxcc_centered), sum(nxcc_offcenter),
                               places=4, msg='Center of rotation for mask is incorrect.')
        self.assertTrue(sum(nxcc_centered) > 99.09, msg='Precision of mask rotation is too low.')



import numpy as np
import unittest
from pytom_tm.weights import (create_wedge, create_ctf, create_gaussian_band_pass, radial_reduced_grid,
                              radial_average, power_spectrum_profile, profile_to_weighting)
from pytom_tm.io import write_mrc


# Dose and ctf params for tomo_104
CS = 2.7
AMP = 0.08
VOL = 200
IMOD_CTF = '''1	1	-50.99	-50.99	  3083   2
2	2	-47.99	-47.99	  3084
3	3	-44.99	-44.99	  3147
4	4	-41.99	-41.99	  3083
5	5	-38.99	-38.99	  3164
6	6	-35.99	-35.99	  3087
7	7	-32.99	-32.99	  3116
8	8	-29.99	-29.99	  3071
9	9	-26.99	-26.99	  3089
10	10	-23.99	-23.99	  3078
11	11	-20.99	-20.99	  3082
12	12	-17.99	-17.99	  3007
13	13	-14.99	-14.99	  3047
14	14	-11.99	-11.99	  3089
15	15	-8.99	-8.99	  3055
16	16	-5.99	-5.99	  3133
17	17	-2.99	-2.99	  3206
18	18	 0.01	 0.01	  3323
19	19	 3.01	 3.01	  3233
20	20	 6.01	 6.01	  3166
21	21	 9.01	 9.01	  3181
22	22	12.01	12.01	  3167
23	23	15.00	15.00	  3183
24	24	18.00	18.00	  3194
25	25	21.01	21.01	  3257
26	26	24.01	24.01	  3281
27	27	27.00	27.00	  3274
28	28	30.01	30.01	  3249
29	29	33.00	33.00	  3326
30	30	36.00	36.00	  3319
31	31	39.00	39.00	  3338
32	32	42.00	42.00	  3309
33	33	45.00	45.00	  3326
34	34	48.00	48.00	  3274
35	35	51.00	51.00	  3354
'''
CTF_PARAMS = []
for d in [float(x.split()[4]) * 1e-3 for x in IMOD_CTF.split('\n') if x != '']:
    CTF_PARAMS.append({
        'defocus': d,
        'amplitude': AMP,
        'voltage': VOL,
        'cs': CS
    })

DOSE_FILE = '''60.165 
58.255 
56.345 
46.795 
44.885 
42.975 
41.065 
35.335 
33.425 
27.695 
25.785 
20.055 
18.145 
12.415 
10.505 
4.775 
2.865 
0.955 
6.685 
8.595 
14.325 
16.235 
21.965 
23.875 
29.605 
31.515 
37.245 
39.155 
48.705 
50.615 
52.525 
54.435 
62.075 
63.985 
65.895 
'''
ACCUMULATED_DOSE = [float(x.strip()) for x in DOSE_FILE.split('\n') if x != '']
TILT_ANGLES = list(range(-51, 54, 3))


class TestWeights(unittest.TestCase):
    def setUp(self):
        self.volume_shape_even = (10, 10, 10)
        self.volume_shape_uneven = (11, 11, 11)
        self.volume_shape_irregular = (7, 12, 6)
        self.voxel_size = 3.34
        self.low_pass = 10
        self.high_pass = 50

        self.reduced_even_shape_3d = (10, 10, 6)
        self.reduced_even_shape_2d = (10, 6)
        self.reduced_uneven_shape_3d = (11, 11, 6)
        self.reduced_uneven_shape_2d = (11, 6)
        self.reduced_irregular_shape_3d = (7, 12, 6 // 2 + 1)
        self.reduced_irregular_shape_2d = (7, 12 // 2 + 1)

    def test_radial_reduced_grid(self):
        with self.assertRaises(ValueError, msg='Radial reduced grid should raise ValueError if the shape is '
                                               'not 2- or 3-dimensional.'):
            radial_reduced_grid((5, ))
        with self.assertRaises(ValueError, msg='Radial reduced grid should raise ValueError if the shape is '
                                               'not 2- or 3-dimensional.'):
 
            radial_reduced_grid((5, ) * 4)

        self.assertEqual(radial_reduced_grid(self.volume_shape_even).shape, self.reduced_even_shape_3d,
                         msg='3D radial reduced grid does not have the correct shape')
        self.assertEqual(radial_reduced_grid(self.volume_shape_even[:2]).shape, self.reduced_even_shape_2d,
                         msg='2D radial reduced grid does not have the correct shape')

    def test_band_pass(self):
        with self.assertRaises(ValueError, msg='Bandpass should raise ValueError if both low and high pass are None'):
            create_gaussian_band_pass(
                self.volume_shape_even,
                self.voxel_size,
                None,
                None
            )
        with self.assertRaises(ValueError, msg='Bandpass should raise ValueError if low pass resolution > high pass '
                                               'resolution'):
            create_gaussian_band_pass(
                self.volume_shape_even,
                self.voxel_size,
                50,
                10
            )
        band_pass = create_gaussian_band_pass(self.volume_shape_even, self.voxel_size, self.low_pass, self.high_pass)
        low_pass = create_gaussian_band_pass(self.volume_shape_even, self.voxel_size, low_pass=self.low_pass)
        high_pass = create_gaussian_band_pass(self.volume_shape_even, self.voxel_size, high_pass=self.high_pass)

        self.assertEqual(band_pass.shape, self.reduced_even_shape_3d,
                         msg='Bandpass filter does not have expected output shape')
        self.assertEqual(band_pass.dtype, np.float64,
                         msg='Bandpass filter does not have expected dtype')
        self.assertEqual(low_pass.shape, self.reduced_even_shape_3d,
                         msg='Low-pass filter does not have expected output shape')
        self.assertEqual(low_pass.dtype, np.float64,
                         msg='Low-pass filter does not have expected dtype')
        self.assertEqual(high_pass.shape, self.reduced_even_shape_3d,
                         msg='High-pass filter does not have expected output shape')
        self.assertEqual(high_pass.dtype, np.float64,
                         msg='High-pass filter does not have expected dtype')

        self.assertTrue(np.sum((band_pass != low_pass) * 1) != 0,
                        msg='Band-pass and low-pass should be different')
        self.assertTrue(np.sum((band_pass != high_pass) * 1) != 0,
                        msg='Band-pass and low-pass filter should be different')
        self.assertTrue(np.sum((low_pass != high_pass) * 1) != 0,
                        msg='Low-pass and high-pass filter should be different')

    def test_create_wedge(self):
        with self.assertRaises(ValueError, msg='Create wedge should raise ValueError if tilt_angles list does not '
                                               'contain at least two values'):
            create_wedge(
                self.volume_shape_even,
                [1.],
                1.
            )
        with self.assertRaises(ValueError, msg='Create wedge should raise ValueError if tilt_angles input is not a '
                                               'list'):
            create_wedge(
                self.volume_shape_even,
                1.,
                1.
            )
        with self.assertRaises(ValueError, msg='Create wedge should raise ValueError if voxel_size is smaller or '
                                               'equal to 0'):
            create_wedge(
                self.volume_shape_even,
                TILT_ANGLES,
                0.
            )
        with self.assertRaises(ValueError, msg='Create wedge should raise ValueError if cut_off_radius is smaller or '
                                               'equal to 0'):
            create_wedge(
                self.volume_shape_even,
                TILT_ANGLES,
                1.,
                cut_off_radius=0.
            )

        # create test wedges
        structured_wedge = create_wedge(self.volume_shape_even, TILT_ANGLES, 1., tilt_weighting=True)
        symmetric_wedge = create_wedge(self.volume_shape_even, [TILT_ANGLES[0], TILT_ANGLES[-1]],
                                       1., tilt_weighting=False)
        asymmetric_wedge = create_wedge(self.volume_shape_even, [TILT_ANGLES[0], TILT_ANGLES[-2]],
                                        1., tilt_weighting=False)

        self.assertEqual(structured_wedge.shape, self.reduced_even_shape_3d,
                         msg='Structured wedge does not have expected output shape')
        self.assertEqual(structured_wedge.dtype, np.float32,
                         msg='Structured wedge does not have expected dtype')

        self.assertEqual(symmetric_wedge.shape, self.reduced_even_shape_3d,
                         msg='Symmetric wedge does not have expected output shape')
        self.assertEqual(symmetric_wedge.dtype, np.float32,
                         msg='Symmetric wedge does not have expected dtype')

        self.assertEqual(asymmetric_wedge.shape, self.reduced_even_shape_3d,
                         msg='Asymmetric wedge does not have expected output shape')
        self.assertEqual(asymmetric_wedge.dtype, np.float32,
                         msg='Asymmetric wedge does not have expected dtype')

        self.assertTrue(np.sum((symmetric_wedge != asymmetric_wedge) * 1) != 0,
                        msg='Symmetric and asymmetric wedge should be different!')

        structured_wedge = create_wedge(self.volume_shape_even, TILT_ANGLES, self.voxel_size, tilt_weighting=True,
                                        cut_off_radius=1., low_pass=self.low_pass, high_pass=self.high_pass)
        self.assertEqual(structured_wedge.shape, self.reduced_even_shape_3d,
                         msg='Wedge with band-pass does not have expected output shape')
        self.assertEqual(structured_wedge.dtype, np.float32,
                         msg='Wedge with band-pass does not have expected dtype')

        # test shapes of wedges
        weights = create_wedge(self.volume_shape_even, TILT_ANGLES, self.voxel_size * 3,
                               tilt_weighting=True, low_pass=40,
                               accumulated_dose_per_tilt=ACCUMULATED_DOSE,
                               ctf_params_per_tilt=CTF_PARAMS)
        self.assertEqual(weights.shape, self.reduced_even_shape_3d,
                         msg='3D CTF does not have the correct reduced fourier shape.')
        weights = create_wedge(self.volume_shape_uneven, TILT_ANGLES, self.voxel_size * 3,
                               tilt_weighting=True, low_pass=40,
                               accumulated_dose_per_tilt=ACCUMULATED_DOSE,
                               ctf_params_per_tilt=CTF_PARAMS)
        self.assertEqual(weights.shape, self.reduced_uneven_shape_3d,
                         msg='3D CTF does not have the correct reduced fourier shape.')

        # test parameter flexibility of tilt_weighted wedge
        weights = create_wedge(self.volume_shape_even, TILT_ANGLES, self.voxel_size * 3,
                               tilt_weighting=True, low_pass=self.low_pass,
                               accumulated_dose_per_tilt=None,
                               ctf_params_per_tilt=None)
        self.assertEqual(weights.shape, self.reduced_even_shape_3d,
                         msg='Tilt weighted wedge should also work without defocus and dose info.')

    def test_ctf(self):
        ctf_raw = create_ctf(
            self.volume_shape_even,
            self.voxel_size * 1E-10,
            3E-6,
            0.08,
            300E3,
            2.7E-3
        )
        ctf_cut = create_ctf(
            self.volume_shape_even,
            self.voxel_size * 1E-10,
            3E-6,
            0.08,
            300E3,
            2.7E-3,
            cut_after_first_zero=True
        )
        self.assertEqual(ctf_raw.shape, self.reduced_even_shape_3d,
                         msg='CTF does not have expected output shape')
        self.assertTrue(np.sum((ctf_raw != ctf_cut) * 1) != 0,
                        msg='CTF should be different when cutting it off after the first zero crossing')

    def test_radial_average(self):
        x, y = 100, 50
        with self.assertRaises(ValueError, msg='Radial average should raise error if something other than 2d/3d '
                                               'array is provided.'):
            radial_average(
                np.zeros(x)
            )
        q, m = radial_average(np.zeros((x, y)))
        self.assertEqual(m.shape[0], x // 2 + 1, msg='Radial average shape should equal largest sampling dimension.')
        q, m = radial_average(np.zeros((30, y)))
        self.assertEqual(m.shape[0], y, msg='Radial average shape should equal largest sampling dimension, '
                                            'considering Fourier reduced form.')
        q, m = radial_average(np.zeros((20, x, y)))
        self.assertEqual(m.shape[0], x // 2 + 1, msg='Radial average shape should equal largest sampling dimension.')
        q, m = radial_average(np.zeros((20, 30, y)))
        self.assertEqual(m.shape[0], y, msg='Radial average shape should equal largest sampling dimension, '
                                            'considering Fourier reduced form.')

    def test_power_spectrum_profile(self):
        with self.assertRaises(ValueError, msg='Power spectrum profile should raise ValueError if input image is '
                                               'not 2- or 3-dimensional.'):
            power_spectrum_profile(np.zeros(5))
        with self.assertRaises(ValueError, msg='Power spectrum profile should raise ValueError if input image is '
                                               'not 2- or 3-dimensional.'):

            power_spectrum_profile(np.zeros((5, ) * 4))
        profile = power_spectrum_profile(np.zeros(self.volume_shape_irregular))
        self.assertEqual(profile.shape, (max(self.volume_shape_irregular) // 2 + 1, ),
                         msg='Power spectrum profile output shape should be a 1-dimensional array with '
                             'length equal to max(input_shape) // 2 + 1, corresponding to largest sampling component '
                             'in Fourier space.')

    def test_profile_to_weighting(self):
        with self.assertRaises(ValueError, msg='Profile to weighting should raise a ValueError if the profile is not '
                                               '1-dimensional.'):
            profile_to_weighting(np.zeros((5, 5)), (5, 5))
        with self.assertRaises(ValueError, msg='Profile to weighting should raise a ValueError if the output shape '
                                               'for the weighting is not 2- or 3-dimensional.'):
            profile_to_weighting(np.zeros(5), (5, ))
        with self.assertRaises(ValueError, msg='Profile to weighting should raise a ValueError if the output shape '
                                               'for the weighting is not 2- or 3-dimensional.'):
            profile_to_weighting(np.zeros(5), (5,) * 4)

        profile = power_spectrum_profile(np.zeros(self.volume_shape_irregular))
        self.assertEqual(profile_to_weighting(profile, self.volume_shape_irregular).shape,
                         self.reduced_irregular_shape_3d,
                         msg='Profile to weighting should return 3D Fourier reduced array.')
        self.assertEqual(profile_to_weighting(profile, self.volume_shape_irregular[:2]).shape,
                         self.reduced_irregular_shape_2d,
                         msg='Profile to weighting should return 2D Fourier reduced array.')



import numpy as np
import numpy.typing as npt
import logging
import scipy.ndimage as ndimage
from typing import Optional, Union
from pytom_tm.io import UnequalSpacingError


constants = {
    # Dictionary of physical constants required for calculation.
    "c": 299792458,  # m/s
    "el": 1.60217646e-19,  # C
    "h": 6.62606896e-34,  # J*S
    "h_ev": 4.13566733e-15,  # eV*s
    "h_bar": 1.054571628e-34,  # J*s
    "h_bar_ev": 6.58211899e-16,  # eV*s

    "na": 6.02214179e23,  # mol-1
    "re": 2.817940289458e-15,  # m
    "rw": 2.976e-10,  # m

    "me": 9.10938215e-31,  # kg
    "me_ev": 0.510998910e6,  # ev/c^2
    "kb": 1.3806503e-23,  # m^2 kgs^-2 K^-1

    "eps0": 8.854187817620e-12  # F/m
}


def hwhm_to_sigma(hwhm: float) -> float:
    """Convert half width of half maximum of a Gaussian to sigma by dividing by sqrt(2 * ln(2)).

    Parameters
    ----------
    hwhm: float
        half width of half maximum of Gaussian

    Returns
    -------
    sigma: float
        sigma of Gaussian
    """
    return hwhm / (np.sqrt(2 * np.log(2)))


def sigma_to_hwhm(sigma: float) -> float:
    """Convert sigma to half width of half maximum of a Gaussian by multiplying with sqrt(2 * ln(2)).

    Parameters
    ----------
    sigma: float
        sigma of Gaussian

    Returns
    -------
    hwhm: float
        half width of half maximum of Gaussian
    """
    return sigma * (np.sqrt(2 * np.log(2)))


def wavelength_ev2m(voltage: float) -> float:
    """Calculate wavelength of electrons from voltage.

    Parameters
    ----------
    voltage: float
        voltage of wave in eV

    Returns
    -------
    lambda: float
        wavelength of electrons in m
    """
    h = constants["h"]
    e = constants["el"]
    m = constants["me"]
    c = constants["c"]

    _lambda = h / np.sqrt(e * voltage * m * (e / m * voltage / c ** 2 + 2))

    return _lambda


def radial_reduced_grid(
        shape: Union[tuple[int, int, int], tuple[int, int]],
        shape_is_reduced: bool = False
) -> npt.NDArray[float]:
    """Calculates a Fourier space radial reduced grid for the given input shape, with the 0 frequency in the center
    of the output image. Values range from 0 in the center to 1 at Nyquist frequency.

    By default, it is assumed shape belongs to a real space array, which causes the function to return a
    grid with the last dimension reduced, i.e. shape[-1] // 2 + 1 (ideal for creating frequency
    dependent filters). However, setting radial_reduced_grid(..., shape_is_reduced=True) the shape is assumed to
    already be in a reduced form.

    Parameters
    ----------
    shape: Union[tuple[int, int, int], tuple[int, int]]
        2D/3D input shape, usually the .shape attribute of a numpy array
    shape_is_reduced: bool, default False
        whether the shape is already in a reduced fourier format, False by default

    Returns
    ----------
    radial_reduced_grid: npt.NDArray[float]
        fourier space frequency grid, 0 in center, 1 at nyquist
    """
    if not len(shape) in [2, 3]:
        raise ValueError('radial_reduced_grid() only works for 2D or 3D shapes')
    reduced_dim = shape[-1] if shape_is_reduced else shape[-1] // 2 + 1
    if len(shape) == 3:
        x = (np.abs(np.arange(
            -shape[0] // 2 + shape[0] % 2,
            shape[0] // 2 + shape[0] % 2, 1.
        )) / (shape[0] // 2))[:, np.newaxis, np.newaxis]
        y = (np.abs(np.arange(
            -shape[1] // 2 + shape[1] % 2,
            shape[1] // 2 + shape[1] % 2, 1.
        )) / (shape[1] // 2))[:, np.newaxis]
        z = np.arange(0, reduced_dim, 1.) / (reduced_dim - 1)
        return np.sqrt(x ** 2 + y ** 2 + z ** 2)
    elif len(shape) == 2:
        x = (np.abs(np.arange(
            -shape[0] // 2 + shape[0] % 2,
            shape[0] // 2 + shape[0] % 2, 1.
        )) / (shape[0] // 2))[:, np.newaxis]
        y = np.arange(0, reduced_dim, 1.) / (reduced_dim - 1)
        return np.sqrt(x ** 2 + y ** 2)


def create_gaussian_low_pass(
        shape: Union[tuple[int, int, int], tuple[int, int]],
        spacing: float,
        resolution: float
) -> npt.NDArray[float]:
    """Create a 3D Gaussian low-pass filter with cutoff (or HWHM) that is reduced in fourier space.

    Parameters
    ----------
    shape: Union[tuple[int, int, int], tuple[int, int]]
        shape tuple with x,y or x,y,z dimension
    spacing: float
        voxel size in real space
    resolution: float
        resolution in real space to filter towards

    Returns
    ----------
    output: npt.NDArray[float]
        array containing the filter
    """
    q = radial_reduced_grid(shape)

    # 2 * spacing / resolution is cutoff in fourier space
    # then convert cutoff (hwhm) to sigma for gaussian function
    sigma_fourier = hwhm_to_sigma(2 * spacing / resolution)

    return np.fft.ifftshift(np.exp(-q ** 2 / (2 * sigma_fourier ** 2)), axes=(0, 1))


def create_gaussian_high_pass(
        shape: Union[tuple[int, int, int], tuple[int, int]],
        spacing: float,
        resolution: float
) -> npt.NDArray[float]:
    """Create a 3D Gaussian high-pass filter with cutoff (or HWHM) that is reduced in fourier space.

    Parameters
    ----------
    shape: Union[tuple[int, int, int], tuple[int, int]]
        shape tuple with x,y or x,y,z dimension
    spacing: float
        voxel size in real space
    resolution: float
        resolution in real space to filter towards

    Returns
    ----------
    output: npt.NDArray[float]
        array containing the filter
    """
    q = radial_reduced_grid(shape)

    # 2 * spacing / resolution is cutoff in fourier space
    # then convert cutoff (hwhm) to sigma for gaussian function
    sigma_fourier = hwhm_to_sigma(2 * spacing / resolution)

    return np.fft.ifftshift(1 - np.exp(-q ** 2 / (2 * sigma_fourier ** 2)), axes=(0, 1))


def create_gaussian_band_pass(
        shape: Union[tuple[int, int, int], tuple[int, int]],
        spacing: float,
        low_pass: Optional[float] = None,
        high_pass: Optional[float] = None
) -> npt.NDArray[float]:
    """Resolution bands presents the resolution shells where information needs to be maintained. For example the bands
    might be (150A, 40A). For a spacing of 15A (nyquist resolution is 30A) this is a mild low pass filter. However,
    quite some low spatial frequencies will be cut by it.

    Parameters
    ----------
    shape: Union[tuple[int, int, int], tuple[int, int]]
        shape tuple with x,y or x,y,z dimension
    spacing: float
        voxel size in real space
    low_pass: Optional[float], default None
        resolution of low-pass filter
    high_pass: Optional[float], default None
        resolution of high-pass filter

    Returns
    ----------
    output: npt.NDArray[float]
        array containing the band-pass filter
    """
    if high_pass is None and low_pass is None:
        raise ValueError('Either low-pass or high-pass needs to be set for band-pass')

    if high_pass is None:
        return create_gaussian_low_pass(shape, spacing, low_pass)
    elif low_pass is None:
        return create_gaussian_high_pass(shape, spacing, high_pass)
    elif low_pass >= high_pass:
        raise ValueError('Second value of band-pass needs to be a high resolution shell.')
    else:
        q = radial_reduced_grid(shape)

        # 2 * spacing / resolution is cutoff in fourier space
        # then convert cutoff (hwhm) to sigma for gaussian function
        sigma_high_pass = hwhm_to_sigma(2 * spacing / high_pass)
        sigma_low_pass = hwhm_to_sigma(2 * spacing / low_pass)

        return np.fft.ifftshift(
            (1 - np.exp(-q ** 2 / (2 * sigma_high_pass ** 2))) * np.exp(-q ** 2 / (2 * sigma_low_pass ** 2)),
            axes=(0, 1)
        )


def create_wedge(
        shape: tuple[int, int, int],
        tilt_angles: list[float, ...],
        voxel_size: float,
        cut_off_radius: float = 1.,
        angles_in_degrees: bool = True,
        low_pass: Optional[float] = None,
        high_pass: Optional[float] = None,
        tilt_weighting: bool = False,
        accumulated_dose_per_tilt: Optional[list[float, ...]] = None,
        ctf_params_per_tilt: Optional[list[dict]] = None
) -> npt.NDArray[float]:
    """This function returns a wedge volume that is either symmetric or asymmetric depending on wedge angle input.

    Parameters
    ----------
    shape: tuple[int, int, int]
        real space shape of volume to which it needs to be applied
    tilt_angles: list[float, ...]
        tilt angles used for reconstructing the tomogram
    voxel_size: float
        voxel size is needed for the calculation of various filters
    cut_off_radius: float, default 1.
        cutoff as a fraction of nyquist, i.e. 1.0 means all the way to nyquist
    angles_in_degrees: bool, default True
        whether angles are in degrees or radians units
    low_pass: Optional[float], default None
        low pass filter resolution in A
    high_pass: Optional[float], default None
        high pass filter resolution in A
    tilt_weighting: bool, default False
        apply tilt weighting
    accumulated_dose_per_tilt: Optional[list[float, ...]], default None
        accumulated dose for each tilt for dose weighting
    ctf_params_per_tilt: Optional[list[dict]], default None
        ctf parameters for each tilt (see _create_tilt_weighted_wedge() for dict specification)

    Returns
    -------
    wedge: npt.NDArray[float]
        wedge volume that is a reduced fourier space object in z, i.e. shape[2] // 2 + 1
    """
    if not isinstance(tilt_angles, list) or len(tilt_angles) < 2:
        raise ValueError('Wedge generation needs at least a list of two tilt angles.')

    if voxel_size <= 0.:
        raise ValueError('Voxel size in create wedge is smaller or equal to 0, which is an invalid voxel spacing.')

    if cut_off_radius > 1:
        print('Warning: wedge cutoff needs to be defined as a fraction of nyquist 0 < c <= 1. Setting value to 1.0.')
        cut_off_radius = 1.0
    elif cut_off_radius <= 0:
        raise ValueError('Invalid wedge cutoff: needs to be larger than 0')

    if angles_in_degrees:
        tilt_angles_rad = [np.deg2rad(w) for w in tilt_angles]
    else:
        tilt_angles_rad = tilt_angles

    if tilt_weighting:
        wedge = _create_tilt_weighted_wedge(
            shape,
            tilt_angles_rad,
            cut_off_radius,
            voxel_size,
            accumulated_dose_per_tilt=accumulated_dose_per_tilt,
            ctf_params_per_tilt=ctf_params_per_tilt
        ).astype(np.float32)
    else:
        wedge_angles = (np.pi / 2 - np.abs(min(tilt_angles_rad)), np.pi / 2 - np.abs(max(tilt_angles_rad)))
        if np.round(wedge_angles[0], 2) == np.round(wedge_angles[1], 2):
            wedge = _create_symmetric_wedge(
                shape,
                wedge_angles[0],
                cut_off_radius
            ).astype(np.float32)
        else:
            wedge = _create_asymmetric_wedge(
                shape,
                (wedge_angles[0], wedge_angles[1]),
                cut_off_radius
            ).astype(np.float32)

    if not (low_pass is None and high_pass is None):
        return wedge * create_gaussian_band_pass(shape, voxel_size, low_pass, high_pass).astype(np.float32)
    else:
        return wedge


def _create_symmetric_wedge(
        shape: tuple[int, int, int],
        wedge_angle: float,
        cut_off_radius: float
) -> npt.NDArray[float]:
    """This function returns a symmetric wedge object. Function should not be imported, user should call
    create_wedge().

    Parameters
    ----------
    shape: tuple[int, int, int]
        real space shape of volume to which it needs to be applied
    wedge_angle: float
        angle describing symmetric wedge in radians
    cut_off_radius: float
        cutoff as a fraction of nyquist, i.e. 1.0 means all the way to nyquist

    Returns
    ----------
    wedge: npt.NDArray[float]
        wedge volume that is a reduced fourier space object in z, i.e. shape[2] // 2 + 1
    """
    x = (np.abs(np.arange(
        -shape[0] // 2 + shape[0] % 2,
        shape[0] // 2 + shape[0] % 2, 1.
    )) / (shape[0] // 2))[:, np.newaxis]
    z = np.arange(0, shape[2] // 2 + 1, 1.) / (shape[2] // 2)

    # calculate the wedge mask with smooth edges
    wedge_2d = x - np.tan(wedge_angle) * z
    limit = (wedge_2d.max() - wedge_2d.min()) / (2 * min(shape[0], shape[2]) // 2)
    wedge_2d[wedge_2d > limit] = limit
    wedge_2d[wedge_2d < -limit] = -limit
    wedge_2d = (wedge_2d - wedge_2d.min()) / (wedge_2d.max() - wedge_2d.min())
    wedge_2d[shape[0] // 2 + 1, 0] = 1  # ensure that the zero frequency point equals 1

    # duplicate in x
    wedge = np.tile(wedge_2d[:, np.newaxis, :], (1, shape[1], 1))

    wedge[radial_reduced_grid(shape) > cut_off_radius] = 0

    # fourier shift to origin
    return np.fft.ifftshift(wedge, axes=(0, 1))


def _create_asymmetric_wedge(
        shape: tuple[int, int, int],
        wedge_angles: tuple[float, float],
        cut_off_radius: float
) -> npt.NDArray[float]:
    """This function returns an asymmetric wedge object. Function should not be imported, user should call create_wedge().

    Parameters
    ----------
    shape: tuple[int, int, int]
        real space shape of volume to which it needs to be applied
    wedge_angles: tuple[float, float]
        two angles describing asymmetric missing wedge in radians
    cut_off_radius: float
        cutoff as a fraction of nyquist, i.e. 1.0 means all the way to nyquist

    Returns
    ----------
    wedge: npt.NDArray[float]
        wedge volume that is a reduced fourier space object in z, i.e. shape[2] // 2 + 1
    """
    x = (np.abs(np.arange(
        -shape[0] // 2 + shape[0] % 2,
        shape[0] // 2 + shape[0] % 2, 1.
    )) / (shape[0] // 2))[:, np.newaxis]
    z = np.arange(0, shape[2] // 2 + 1, 1.) / (shape[2] // 2)

    # calculate wedge for first angle
    wedge_section = x - np.tan(wedge_angles[0]) * z
    limit = (wedge_section.max() - wedge_section.min()) / (2 * min(shape[0], shape[2]) // 2)
    wedge_section[wedge_section > limit] = limit
    wedge_section[wedge_section < -limit] = -limit
    wedge_section = (wedge_section - wedge_section.min()) / (wedge_section.max() - wedge_section.min())

    # set top of the wedge
    wedge_2d = wedge_section.copy()

    # calculate wedge for second angle
    wedge_section = x - np.tan(wedge_angles[1]) * z
    limit = (wedge_section.max() - wedge_section.min()) / (2 * min(shape[0], shape[2]) // 2)
    wedge_section[wedge_section > limit] = limit
    wedge_section[wedge_section < -limit] = -limit
    wedge_section = (wedge_section - wedge_section.min()) / (wedge_section.max() - wedge_section.min())

    # set bottom of wedge and put 0 frequency to 1
    wedge_2d[shape[0] // 2 + 1:] = wedge_section[shape[0] // 2 + 1:]
    wedge_2d[shape[0] // 2 + 1, 0] = 1

    # duplicate in x
    wedge = np.tile(wedge_2d[:, np.newaxis, :], (1, shape[1], 1))

    wedge[radial_reduced_grid(shape) > cut_off_radius] = 0

    # fourier shift to origin
    return np.fft.ifftshift(wedge, axes=(0, 1))


def _create_tilt_weighted_wedge(
        shape: tuple[int, int, int],
        tilt_angles: list[float, ...],
        cut_off_radius: float,
        pixel_size_angstrom: float,
        accumulated_dose_per_tilt: Optional[list[float, ...]] = None,
        ctf_params_per_tilt: Optional[list[dict]] = None
) -> npt.NDArray[float]:
    """
    The following B-factor heuristic is used (as mentioned in the M paper, and introduced in RELION 1.4):
        "The B factor is increased by 4Å2 per 1e− Å−2 of exposure, and each tilt is weighted as cos θ."

    Relation between B-factor and the sigma of a gaussian:

        B = 8 * pi ** 2 * sigma_motion ** 2

    i.e. sigma_motion = sqrt( B / (8 * pi ** 2)). Belonging to a Gaussian blur:

        exp( -2 * pi ** 2 * sigma_motion ** 2 * q ** 2 )

    Parameters
    ----------
    shape: tuple[int, int, int]
        shape of volume to model the wedge for
    tilt_angles: list[float, ...]
        tilt angles is a list of angle in radian units
    cut_off_radius: float
        cut off for the mask as a fraction of nyquist, value between 0 and 1
    pixel_size_angstrom: float
        the pixel size as a value in Å
    accumulated_dose_per_tilt: list[float, ...], default None
        the accumulated dose in e− Å−2
    ctf_params_per_tilt: list[dict, ...], default None
        the ctf parameters per tilt angle, list of dicts where each dict has the following keys:
        - 'defocus'; in um
        - 'amplitude'; fraction of amplitude contrast between 0 and 1
        - 'voltage'; in keV
        - 'cs'; spherical abberation in mm

    Returns
    -------
    wedge: npt.NDArray[float]
        structured wedge mask in fourier reduced form, i.e. output shape is (shape[0], shape[1], shape[2] // 2 + 1)
    """
    if accumulated_dose_per_tilt is not None and len(accumulated_dose_per_tilt) != len(tilt_angles):
        raise ValueError('in _create_tilt_weighted_wedge the list of accumulated dose per tilt does not have the same '
                         'length as the tilt angle list!')
    if ctf_params_per_tilt is not None and len(ctf_params_per_tilt) != len(tilt_angles):
        raise ValueError('in _create_tilt_weighted_wedge the list of CTF parameters per tilt does not have the same '
                         'length as the tilt angle list!')
    if not all([shape[0] == s for s in shape[1:]]):
        raise UnequalSpacingError('Input shape for structured wedge needs to be a square box. '
                                  'Otherwise the frequencies in fourier space are not equal across dimensions.')

    image_size = shape[0]  # assign to size variable as all dimensions are equal size
    tilt = np.zeros(shape)
    q_grid = radial_reduced_grid(shape)
    tilt_weighted_wedge = np.zeros((image_size, image_size, image_size // 2 + 1))

    for i, alpha in enumerate(tilt_angles):
        if ctf_params_per_tilt is not None:
            ctf = np.fft.fftshift(
                create_ctf(
                    (image_size, ) * 2,
                    pixel_size_angstrom * 1e-10,
                    ctf_params_per_tilt[i]['defocus'] * 1e-6,
                    ctf_params_per_tilt[i]['amplitude'],
                    ctf_params_per_tilt[i]['voltage'] * 1e3,
                    ctf_params_per_tilt[i]['cs'] * 1e-3,
                    flip_phase=True  # creating a per tilt ctf is hard if the phase is not flipped
                ), axes=0,
            )
            tilt[:, :, image_size // 2] = np.concatenate(
                (  # duplicate and flip the CTF around the 0 frequency; then concatenate to make it non-reduced
                    np.flip(ctf[:, 1: 1 + image_size - ctf.shape[1]], axis=1),
                    ctf
                ),
                axis=1
            )
        else:
            tilt[:, :, image_size // 2] = 1

        # rotate the image weights to the tilt angle
        rotated = np.flip(
            ndimage.rotate(
                tilt,
                np.rad2deg(alpha),
                axes=(0, 2),
                reshape=False,
                order=3
            )[:, :, : image_size // 2 + 1]
        )

        # weight with exposure and tilt dampening
        if accumulated_dose_per_tilt is not None:
            q_squared = (q_grid / (2 * pixel_size_angstrom)) ** 2
            sigma_motion = np.sqrt(accumulated_dose_per_tilt[i] * 4 / (8 * np.pi ** 2))
            weighted_tilt = (
                    rotated *
                    np.cos(alpha) *  # apply tilt-dependent weighting
                    np.exp(-2 * np.pi ** 2 * sigma_motion ** 2 * q_squared)  # apply dose-weighting
            )
        else:
            weighted_tilt = (
                    rotated *
                    np.cos(alpha)  # apply tilt-dependent weighting
            )
        tilt_weighted_wedge = np.maximum(tilt_weighted_wedge, weighted_tilt)

    tilt_weighted_wedge[q_grid > cut_off_radius] = 0

    return np.fft.ifftshift(tilt_weighted_wedge, axes=(0, 1))


def create_ctf(
        shape: Union[tuple[int, int, int], tuple[int, int]],
        pixel_size: float,
        defocus: float,
        amplitude_contrast: float,
        voltage: float,
        spherical_aberration: float,
        cut_after_first_zero: bool = False,
        flip_phase: bool = False
) -> npt.NDArray[float]:
    """Create a CTF in a 3D volume in reduced format.

    Parameters
    ----------
    shape: Union[tuple[int, int, int], tuple[int, int]]
        dimensions of volume to create ctf in
    pixel_size: float
        pixel size for ctf in m
    defocus: float
        defocus for ctf in m
    amplitude_contrast: float
        the fraction of amplitude contrast in the ctf
    voltage: float
        acceleration voltage of the microscope in eV
    spherical_aberration: float
        spherical aberration in m
    cut_after_first_zero: bool, default False
        whether to cut ctf after first zero crossing
    flip_phase: bool, default False
        make ctf fully positive/negative to imitate ctf correction by phase flipping

    Returns
    -------
    ctf: npt.NDArray[float]
        CTF in 3D
    """
    k = radial_reduced_grid(shape) / (2 * pixel_size)  # frequencies in fourier space

    _lambda = wavelength_ev2m(voltage)

    # phase contrast transfer
    chi = np.pi * _lambda * defocus * k ** 2 - 0.5 * np.pi * spherical_aberration * _lambda ** 3 * k ** 4
    # amplitude contrast term
    tan_term = np.arctan(amplitude_contrast / np.sqrt(1 - amplitude_contrast ** 2))

    # determine the ctf
    ctf = - np.sin(chi + tan_term)

    if cut_after_first_zero:  # find frequency to cut first zero
        def chi_1d(q):
            return np.pi * _lambda * defocus * q ** 2 - 0.5 * np.pi * spherical_aberration * _lambda ** 3 * q ** 4

        def ctf_1d(q):
            return - np.sin(chi_1d(q) + tan_term)

        # sample 1d ctf and get indices of zero crossing
        k_range = np.arange(max(k.shape)) / max(k.shape) / (2 * pixel_size)
        values = ctf_1d(k_range)
        zero_crossings = np.where(np.diff(np.sign(values)))[0]

        # for overfocus the first crossing needs to be skipped, for example see: Yonekura et al. 2006 JSB
        k_cutoff = k_range[zero_crossings[0]] if defocus > 0 else k_range[zero_crossings[1]]

        # filter the ctf with the cutoff frequency
        ctf[k > k_cutoff] = 0

    if flip_phase:  # phase flip
        ctf = np.abs(ctf)

    return np.fft.ifftshift(ctf, axes=(0, 1) if len(shape) == 3 else 0)


def radial_average(weights: npt.NDArray[float]) -> tuple[npt.NDArray[float], npt.NDArray[float]]:
    """ This calculates the radial average of a reduced fourier space function.

    Parameters
    ----------
    weights: npt.NDArray[float]
        3D array to be radially averaged: in fourier reduced form and with the origin in the corner.

    Returns
    -------
    (q, mean): tuple[npt.NDArray[float], npt.NDArray[float]]
        A tuple of two 1d numpy arrays. Their length equals half of largest input dimension.
    """
    if len(weights.shape) not in [2, 3]:
        raise ValueError('Radial average calculation only works for 2d/3d arrays')

    # get the number of sampling points from the largest fourier dimension,
    # unless the reduced dimensions is already the largest one
    sampling_points = max(
        max(weights.shape[:-1]) // 2 + 1,
        weights.shape[-1]
    )
                           
    q = np.arange(sampling_points)
    q_grid = np.floor(
        # convert to radial indices in the fourier power spectrum, 0.5 is added to obtain the correct ring
        radial_reduced_grid(weights.shape, shape_is_reduced=True) * (sampling_points - 1) + 0.5
    ).astype(int)
    mean = ndimage.mean(
        np.fft.fftshift(weights, axes=(0, 1) if len(weights.shape) == 3 else 0),
        labels=q_grid,
        index=q
    )

    return q, mean


def power_spectrum_profile(image: npt.NDArray[float]) -> npt.NDArray[float]:
    """ Calculate the power spectrum for a real space array and then find the profile (radial average).

    Parameters
    ----------
    image: npt.NDArray[float]
        2D/3D array in real space for which the power spectrum profile needs to be calculated

    Returns
    -------
    profile: npt.NDArray[float]
        A 1D numpy array
    """
    if len(image.shape) not in [2, 3]:
        raise ValueError('Power spectrum profile calculation only works for 2d/3d arrays.')

    _, power_profile = radial_average(np.abs(np.fft.rfftn(image)) ** 2)

    return power_profile


def profile_to_weighting(
        profile: npt.NDArray[float],
        shape: Union[tuple[int, int], tuple[int, int, int]]
) -> npt.NDArray[float]:
    """ Calculate a radial weighing (filter) from a spectrum profile.

    Parameters
    ----------
    profile: npt.NDArray[float]
        power spectrum profile (or other 1d profile) to transform in a fourier space filter
    shape: Union[tuple[int, int], tuple[int, int, int]]
        2D/3D array shape in real space for which the fourier reduced weights are calculated

    Returns
    -------
    weighting: npt.NDArray[float]
        Reduced Fourier space weighting for shape
    """
    if len(profile.shape) != 1:
        raise ValueError('Profile passed to profile_to_weighting is not 1-dimensional.')
    if len(shape) not in [2, 3]:
        raise ValueError('Shape passed to profile_to_weighting needs to be 2D/3D.')

    q_grid = radial_reduced_grid(shape)

    weights = ndimage.map_coordinates(
        profile,
        q_grid.flatten()[np.newaxis, :] * profile.shape[0],
        order=1
    ).reshape(q_grid.shape)

    weights[q_grid > 1] = 0

    return np.fft.ifftshift(weights, axes=(0, 1) if len(shape) == 3 else 0)
